package logrusloki

import (
	"fmt"
	"io"
	"os"
	"path/filepath"
	"strconv"
	"time"

	"github.com/cortexproject/cortex/pkg/util"
	"github.com/cortexproject/cortex/pkg/util/flagext"
	"github.com/go-kit/kit/log"
	"github.com/go-kit/kit/log/level"
	klogrus "github.com/go-kit/kit/log/logrus"
	"github.com/grafana/loki/pkg/promtail/client"
	"github.com/prometheus/common/config"
	"github.com/prometheus/common/model"
	"github.com/sirupsen/logrus"
)

type Hook interface {
	logrus.Hook
	io.Closer
}

type hook struct {
	client client.Client
	log    log.Logger

	level logrus.Level

	app string

	fieldsAsLabels bool
	labels         model.LabelSet
}

var (
	defaultClientConfig = client.Config{
		BatchWait: 1 * time.Second,
		BatchSize: 10 * 1024,
		BackoffConfig: util.BackoffConfig{
			MinBackoff: 100 * time.Millisecond,
			MaxBackoff: 10 * time.Second,
			MaxRetries: 10,
		},
		Timeout: 10 * time.Second,
	}
)

func NewHookFromEnv(app string, level logrus.Level, opts ...Option) (Hook, error) {
	host := os.Getenv("LOKI_ADDR")
	user := os.Getenv("LOKI_USERNAME")
	pass := os.Getenv("LOKI_PASSWORD")
	return NewHook(app, host, user, pass, level, opts...)
}

func NewHook(app, host, user, pass string, level logrus.Level, opts ...Option) (Hook, error) {
	url := flagext.URLValue{}
	if err := url.Set(host); err != nil {
		return nil, err
	}
	url.Path = filepath.Join(url.Path, "/loki/api/v1/push")
	c := defaultClientConfig
	c.URL = url
	c.Client = config.HTTPClientConfig{
		BasicAuth: &config.BasicAuth{
			Username: user,
			Password: config.Secret(pass),
		},
	}
	return NewHookFromConfig(app, c, level, opts...)
}

func NewHookFromConfig(app string, config client.Config, level logrus.Level, opts ...Option) (Hook, error) {
	klog := klogrus.NewLogrusLogger(logrus.New())
	c, err := client.New(config, klog)
	if err != nil {
		return nil, err
	}
	labels := make(model.LabelSet)
	labels["app"] = model.LabelValue(app)

	log := log.With(klog, "component", "logrusLoki")
	h := &hook{client: c, level: level, app: app, labels: labels, log: log}
	for _, o := range opts {
		o(h)
	}
	return h, nil
}

func (h *hook) Levels() []logrus.Level {
	return logrus.AllLevels[:h.level+1]
}

func (h *hook) Fire(e *logrus.Entry) error {
	msg := e.Message
	labels := h.labels.Clone()
	if h.fieldsAsLabels {
		for k, v := range e.Data {
			labels[model.LabelName(k)] = model.LabelValue(fmt.Sprintf("%v", v))
		}
		labels["level"] = model.LabelValue(e.Level.String())
		if e.Caller != nil {
			labels["go_function"] = model.LabelValue(e.Caller.Function)
			labels["go_file"] = model.LabelValue(e.Caller.File)
			labels["go_line"] = model.LabelValue(strconv.Itoa(e.Caller.Line))
		}
	} else {
		b, err := (&logrus.TextFormatter{DisableColors: true}).Format(e)
		if err != nil {
			level.Error(h.log).Log("msg", "failed to format error", "error", err.Error())
		} else {
			msg = string(b)
		}
	}
	go func() {
		if err := h.client.Handle(labels, e.Time, msg); err != nil {
			level.Error(h.log).Log("msg", "failed to send logs", "error", err.Error())
		}
	}()
	return nil
}

func (h *hook) Close() error {
	h.client.Stop()
	return nil
}
